#!/opt/alt/python37/bin/python3.7 -sbb
# -*- coding: utf-8 -*-

# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENSE.TXT

from __future__ import absolute_import

import json
import os
import sys
from typing import Set

from clcommon.clpwd import drop_privileges
from clcommon.cpapi import docroot
from clwpos.utils import WposUser, get_parent_pid
from cpanel_user_disable_caching import (
    create_php_info_file,
    check_domains_wpos_info_and_disable_cache,
    is_object_cache_enabled_for_user,
    remove_php_info_file,
)


def _run_pre_hook(users: Set[WposUser]):
    """
    Create file with data about domains' php versions in user's .clwpos directory.
    """
    for user in users:
        with drop_privileges(user.name):
            if not is_object_cache_enabled_for_user(user):
                continue
            create_php_info_file(user)


def _run_post_hook(domains: Set[str], users: Set[WposUser]):
    """
    Check incompatibilies for users' domains and disable object caching if needed.

    Sequential processing of users takes quite a long time.
    This method requires optimization or running in background process.
    For now it was not possible to implement launch in background process
    since cPanel parent process tracks all child processes
    (including those called via subprocess.Popen and double os.fork)
    and continues to wait util all its child processes are terminated.
    """
    for user in users:
        with drop_privileges(user.name):
            if not is_object_cache_enabled_for_user(user):
                continue
            check_domains_wpos_info_and_disable_cache(domains)
            remove_php_info_file(user)


def main(args):
    """
    If admin triggers this function we are checking all affected users wps for compatibility
    Need optimization in future
    Hook is triggered by events
      Web - WHM: -> MultiPHP Manager: -> User Domain Settings
      CLI - Whostmgr Lang::PHP::set_vhost_versions post
    """
    domains_affected = {v for k, v in args['data'].items() if 'vhost' in k}
    users_affected = {WposUser(docroot(domain)[1]) for domain in domains_affected}

    hook_type = args['hook']['stage']
    if hook_type == 'pre' and 'version' in args['data']:
        _run_pre_hook(users_affected)
    elif hook_type == 'post':
        _run_post_hook(domains_affected, users_affected)


if __name__ == '__main__':
    argv = json.loads(sys.stdin.read())
    pid = get_parent_pid()
    # set special envars, further when wp-cli will be called,
    # presence of these envars will mean to use PHP version that had been written to the special file
    os.environ.update({'CLWPOS_USE_SAVED_PHP_VERSION': '1', 'CLWPOS_PHP_FILE_ID': str(pid)})
    sys.exit(main(argv))
