#!/opt/alt/python37/bin/python3.7 -sbb
# -*- coding: utf-8 -*-

# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENSE.TXT

from __future__ import absolute_import

import json
import os
import sys
from typing import Dict, Set, Union

from clwpos.cl_wpos_exceptions import WposError
from clwpos.optimization_modules import OBJECT_CACHE_MODULE
from clwpos.user.config import UserConfig
from clwpos.utils import WposUser, get_parent_pid, is_run_under_user, run_in_cagefs_if_needed
from cpanel import php_info as get_php_info


def disable_cache(domain: str, wp_path: str) -> None:
    """
    Disable cache for specified domain and wp_path
    Arguments:
        domain: name of domain ex: domain.com
        wp_path: path to wordpress site
    """
    disable_cmd = ['/usr/bin/clwpos-user', 'disable', f'--domain={domain}',
                   f'--wp-path={wp_path}', f'--module={OBJECT_CACHE_MODULE}']
    run_in_cagefs_if_needed(disable_cmd)


def check_has_incompatibilities(wordpress_info: Dict[str, Union[str, dict]]) -> bool:
    """
    Check for and disable object cache if incompatibility issues found for specified wp
    """
    module_info = wordpress_info.get('modules', {}).get(OBJECT_CACHE_MODULE, {})
    is_enabled = module_info.get('enabled', False)
    issues = module_info.get('issues', [])
    if is_enabled and any(issue['type'] == 'incompatibility' for issue in issues):
        return True
    return False


def create_php_info_file(user: WposUser) -> None:
    """
    Create .php_info-<id> file user's .clwpos directory
    and write data about user's php versions into it.
    """
    if not is_run_under_user():
        raise WposError('Internal Error. Contact CloudLinux support')
    _pid = os.environ.get('CLWPOS_PHP_FILE_ID')
    if not os.path.isdir(user.wpos_dir) or _pid is None:
        return
    php_info = get_php_info()
    file_path = user.php_info.format(file_id=_pid)
    with open(file_path, 'w') as f:
        json.dump(php_info, f)


def remove_php_info_file(user: WposUser) -> None:
    """
    Remove .php_info-<id> file from user's .clwpos directory if it exists.
    """
    if not is_run_under_user():
        raise WposError('Internal Error. Contact CloudLinux support')
    _pid = os.environ.get('CLWPOS_PHP_FILE_ID')
    file_path = user.php_info.format(file_id=_pid)
    if os.path.isfile(file_path):
        os.unlink(file_path)


def is_object_cache_enabled_for_user(user: WposUser) -> bool:
    """
    Return true if object_cache module is enabled for at least one user's site.
    """
    if os.path.exists(user.wpos_dir) \
            and UserConfig(user.name).get_enabled_sites_count_by_module(OBJECT_CACHE_MODULE) > 0:
        return True
    return False


def check_domains_wpos_info_and_disable_cache(domains: Set[str]) -> None:
    """
    Call clwpos-user get command to obtain WPOS data about user's sites.
    If site has incompatibilities, disable object caching on it.
    """
    get_cmd = ['/usr/bin/clwpos-user', 'get']
    result = run_in_cagefs_if_needed(get_cmd)

    for docroot_info in json.loads(result.stdout).get('docroots', []):
        docroot_domains = docroot_info['domains']
        if any(domain in domains for domain in docroot_domains):
            for wordpress_info in docroot_info['wps']:
                if check_has_incompatibilities(wordpress_info):
                    disable_cache(docroot_domains[0], wordpress_info['path'])


def _run_pre_hook(user: WposUser):
    """
    Create file with data about domains' php versions in user's .clwpos directory.
    """
    create_php_info_file(user)


def _run_post_hook(user: WposUser, args: Dict[str, str]):
    """
    Check incompatibilies for user's domains and disable object caching if needed.
    """
    domains_affected = {v for k, v in args.items() if 'vhost' in k}
    check_domains_wpos_info_and_disable_cache(domains_affected)
    remove_php_info_file(user)


def main(args):
    """
    This hook will be triggered in case of user changing php version from cpanel for any of wp
    This module will disable object_cache
    if not supported php_version or php-fpm handler exists in users wordpress sites
    Hook is triggered by events
    Web - Cpanel: -> MultiPHP Manager
    CLI - Cpanel UAPI::LangPHP::php_set_vhost_versions post
    """
    user = WposUser(args['data']['user'])
    if not is_object_cache_enabled_for_user(user):
        return
    hook_type = args['hook']['stage']
    if hook_type == 'pre':
        _run_pre_hook(user)
    elif hook_type == 'post':
        _run_post_hook(user, args['data']['args'])


if __name__ == '__main__':
    argv = json.loads(sys.stdin.read())
    pid = get_parent_pid()
    # set special envars, further when wp-cli will be called,
    # presence of these envars will mean to use PHP version that had been written to the special file
    os.environ.update({'CLWPOS_USE_SAVED_PHP_VERSION': '1', 'CLWPOS_PHP_FILE_ID': str(pid)})
    sys.exit(main(argv))
