#!/opt/alt/python37/bin/python3.7 -bb
# -*- coding: utf-8 -*-

# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENSE.TXT

from __future__ import absolute_import

import argparse
import os
import sys
from typing import List, Optional

from cllimits.cagefs_lib import CageFs
from clwpos.constants import ALT_PHP_PREFIX
from clwpos.optimization_modules import is_module_allowed_for_user
from clwpos.utils import (
    PHP,
    PhpIniConfig,
    get_alt_php_versions,
    is_php_supported_by_wpos,
    is_shared_pro_safely,
    create_pid_file,
    acquire_lock,
)
from cpanel import filter_php_versions_with_not_loaded_redis


def enable_redis_extension(php_version: PHP):
    """
    Enable 'redis' extension for specified alt-php version
    by adding 'extension=redis.so' into php.ini file.
    Also enable 'json' and 'igbinary' extensions if exist which are reqired by 'redis' extension.
    At the same time:
    - disable extension in 'php.d.all/<extension>.ini' to deny user
      to enable extension via PHP Selector.
    - disable extensions in 'php.d/default.ini' to eliminate duplicate loading of extensions.
    """
    php_ini_config = PhpIniConfig(php_version)
    redis_ini_exists = php_ini_config.disable_modules('etc/php.d.all/redis.ini', ['igbinary', 'redis'])
    json_ini_exists = php_ini_config.disable_modules('etc/php.d.all/json.ini', ['json'])
    igbinary_ini_exists = php_ini_config.disable_modules('etc/php.d.all/igbinary.ini', ['igbinary'])

    modules = ['igbinary', 'json', 'redis']
    existing_modules = [
        module
        for module, exists in zip(modules, (igbinary_ini_exists, json_ini_exists, redis_ini_exists))
        if exists
    ]
    php_ini_config.enable_modules('etc/php.ini', existing_modules)
    php_ini_config.disable_modules('etc/php.d/default.ini', existing_modules)


def get_php_versions_to_enable_redis(version: Optional[str]) -> List[PHP]:
    """
    Return list of php version supported by WPOS
    and on which redis is presented but not enabled.
    """
    alt_php_versions = [PHP(f'{ALT_PHP_PREFIX}{version}')] if version else get_alt_php_versions()
    supported_alt_php_versions = [
        version for version in alt_php_versions if is_php_supported_by_wpos(version)
    ]
    return filter_php_versions_with_not_loaded_redis(supported_alt_php_versions)


def parse_args() -> argparse.Namespace:
    """
    Parse arguments for enable_redis_for_alt_php.py utility.
    """
    parser = argparse.ArgumentParser(
        prog='enable_redis_for_alt_php',
        description='Utility for enabling redis extension for alt-php',
    )
    parser.add_argument(
        '--version',
        type=int,
        help='Enable redis extension for specific alt-php version',
    )
    parser.add_argument(
        '--no-rebuild',
        action='store_true',
        help='Run without rebuilding users\' alt_php.ini files',
    )
    opts, _ = parser.parse_known_args()
    return opts


def main():
    """
    Enable redis extension for alt-php versions supported by WPOS.
    """
    if not is_shared_pro_safely(safely=True):
        print('Run only on CL Shared Pro')
        sys.exit(0)
    if not is_module_allowed_for_user():
        print('Run only if WPOS is allowed for at least one user')
        sys.exit(0)

    opts = parse_args()
    cagefs = CageFs()

    with acquire_lock(os.path.join('/var/run', ALT_PHP_PREFIX), attempts=None):
        php_versions_to_enable_redis = get_php_versions_to_enable_redis(opts.version)
        # we don't continue if there are no alt-php versions (supported by WPOS)
        # with redis presented but not loaded
        if not php_versions_to_enable_redis:
            sys.exit(0)

        with create_pid_file(ALT_PHP_PREFIX):
            for php_version in php_versions_to_enable_redis:
                enable_redis_extension(php_version)
            # Rebuild users' alt_php.ini if all condition are satisfied:
            # - script called with no flag --no-rebuild
            # - cagefs package is installed
            if not opts.no_rebuild and cagefs.is_cagefs_present():
                cagefs.rebuild_alt_php_ini()


if __name__ == '__main__':
    main()
